/*
 * Driver for the TAS5766M Audio Amplifier
 *
 * Author: Andy Liu <andy-liu@ti.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/slab.h>
#include <linux/of.h>
#include <linux/init.h>
#include <linux/i2c.h>
#include <linux/regmap.h>
#include <linux/delay.h>

#include <sound/tlv.h>
#include <sound/soc.h>
#include <sound/pcm.h>
#include <sound/initval.h>

#define TAS5766M_DRV_NAME          "tas5766m"

#define TAS5766M_RATES	           (SNDRV_PCM_RATE_48000)
#define TAS5766M_FORMATS           (SNDRV_PCM_FMTBIT_S16_LE | SNDRV_PCM_FMTBIT_S20_3LE |\
			                        SNDRV_PCM_FMTBIT_S24_LE | SNDRV_PCM_FMTBIT_S32_LE)

#define TAS5766M_RQST              (1 << 4)

#define TAS5766M_REG_00            (0x00)
#define TAS5766M_REG_03            (0x03)

#define TAS5766M_VIRT_BASE         (0x100)
#define TAS5766M_PAGE_LEN          (0x100)
#define TAS5766M_PAGE_BASE(n)      (TAS5766M_VIRT_BASE + (TAS5766M_PAGE_LEN * n))

#define TAS5766M_POWER             (TAS5766M_PAGE_BASE(0) +   2)
#define TAS5766M_DIGITAL_VOLUME_2  (TAS5766M_PAGE_BASE(0) +  61)
#define TAS5766M_DIGITAL_VOLUME_3  (TAS5766M_PAGE_BASE(0) +  62)

struct tas5766m_priv {
	struct regmap *regmap;
};

const struct regmap_config tas5766m_regmap = {
	.reg_bits = 8,
	.val_bits = 8,
	.cache_type = REGCACHE_RBTREE,
};

static const DECLARE_TLV_DB_SCALE(digital_tlv, -10350, 50, 1);

static const struct snd_kcontrol_new tas5766m_controls[] = {
SOC_DOUBLE_R_TLV("Digital Playback Volume", TAS5766M_DIGITAL_VOLUME_2,
		 TAS5766M_DIGITAL_VOLUME_3, 0, 255, 1, digital_tlv),
};

static int tas5766m_set_bias_level(struct snd_soc_codec *codec, enum snd_soc_bias_level level)
{
	struct tas5766m_priv *tas5766m = snd_soc_codec_get_drvdata(codec);
	int ret;

	switch (level) {
	case SND_SOC_BIAS_ON:
	case SND_SOC_BIAS_PREPARE:
		break;

	case SND_SOC_BIAS_STANDBY:
		ret = regmap_update_bits(tas5766m->regmap, TAS5766M_POWER, TAS5766M_RQST, 0);
		if (ret != 0) {
			dev_err(codec->dev, "Failed to remove standby: %d\n",
				ret);
			return ret;
		}
		break;

	case SND_SOC_BIAS_OFF:
		ret = regmap_update_bits(tas5766m->regmap, TAS5766M_POWER, TAS5766M_RQST, TAS5766M_RQST);
		if (ret != 0) {
			dev_err(codec->dev, "Failed to request standby: %d\n",
				ret);
			return ret;
		}
		break;
	}

	return 0;
}

static struct snd_soc_codec_driver tas5766m_soc_codec = {
	.set_bias_level = tas5766m_set_bias_level,
	.idle_bias_off = true,
	.controls = tas5766m_controls,
	.num_controls = ARRAY_SIZE(tas5766m_controls),		
};

static int tas5766m_mute(struct snd_soc_dai *dai, int mute)
{
	u8 reg3_value = 0;
	struct snd_soc_codec *codec = dai->codec;

	if (mute)
		reg3_value = 0x11;
	
	snd_soc_write(codec, TAS5766M_REG_00, 0x00);
	snd_soc_write(codec, TAS5766M_REG_03, reg3_value);

	return 0;
}

static const struct snd_soc_dai_ops tas5766m_dai_ops = {
	.digital_mute = tas5766m_mute,
};

static struct snd_soc_dai_driver tas5766m_dai = {
	.name		= "tas5766m-amplifier",
	.playback 	= {
		.stream_name	= "Playback",
		.channels_min	= 2,
		.channels_max	= 2,
		.rates		= TAS5766M_RATES,
		.formats	= TAS5766M_FORMATS,
	},
	.ops = &tas5766m_dai_ops,
};

static int tas5766m_probe(struct device *dev, struct regmap *regmap)
{
	struct tas5766m_priv *tas5766m;
	int ret;
		
	tas5766m = devm_kzalloc(dev, sizeof(struct tas5766m_priv), GFP_KERNEL);
	if (!tas5766m)
		return -ENOMEM;

	dev_set_drvdata(dev, tas5766m);
	tas5766m->regmap = regmap;

	ret = snd_soc_register_codec(dev, 
	                             &tas5766m_soc_codec,
			                     &tas5766m_dai, 
								 1);
	if (ret != 0) 
	{
		dev_err(dev, "Failed to register CODEC: %d\n", ret);
		goto err;
	}

	return 0;
	
err:
	return ret;

}

static int tas5766m_i2c_probe(struct i2c_client *i2c, const struct i2c_device_id *id)
{	
	struct regmap *regmap;
	struct regmap_config config = tas5766m_regmap;
	
	regmap = devm_regmap_init_i2c(i2c, &config);
	if (IS_ERR(regmap))
		return PTR_ERR(regmap);

	return tas5766m_probe(&i2c->dev, regmap);
}

static int tas5766m_remove(struct device *dev)
{
	snd_soc_unregister_codec(dev);
	
	return 0;
}

static int tas5766m_i2c_remove(struct i2c_client *i2c)
{
	tas5766m_remove(&i2c->dev);
	
	return 0;
}

static const struct i2c_device_id tas5766m_i2c_id[] = {
	{ "tas5766m", },
	{ }
};
MODULE_DEVICE_TABLE(i2c, tas5766m_i2c_id);

#ifdef CONFIG_OF
static const struct of_device_id tas5766m_of_match[] = {
	{ .compatible = "ti,tas5766m", },
	{ }
};
MODULE_DEVICE_TABLE(of, tas5766m_of_match);
#endif

static struct i2c_driver tas5766m_i2c_driver = {
	.probe 		= tas5766m_i2c_probe,
	.remove 	= tas5766m_i2c_remove,
	.id_table	= tas5766m_i2c_id,
	.driver		= {
		.name	= TAS5766M_DRV_NAME,
		.of_match_table = tas5766m_of_match,
	},
};

module_i2c_driver(tas5766m_i2c_driver);

MODULE_AUTHOR("Andy Liu <andy-liu@ti.com>");
MODULE_DESCRIPTION("TAS5766M Audio Amplifier Driver - Basic");
MODULE_LICENSE("GPL");
